/* Copyright 1997 Lucent Technologies Inc. - All Rights Reserved */

/* This is an ugly-but-useful test of the crio routines. */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <netdb.h>
#include <netinet/in.h>
#include <signal.h>
#include <errno.h>
#include <string.h>
#include <fcntl.h>
#include "arg.h"

#include "crypto.h"
#include "client.h"
#include "server.h"
#include "stage.h"
#include "util.h"

#define TCP_ECHO	7

#define TMP1	"/tmp/ntest.1"
#define TMP2	"/tmp/ntest.2"

#define SAMPLE	"unstage.c"
/*#define SAMPLE	"/bsd"*/

int debug = 0;

KEY key0, key1, key2, key3;

int servanswer = 0;

char *
setservice(char *servicename, KEY key) {
	int n = strlen(servicename);
	int i;

	switch (servanswer) {
	case 0:
		return "unknown service";
	case 1:
		key1[1] = 0;	/* foul up the works */
		memcpy(key, key1, sizeof(KEY));
		return 0;
	case 2:
		memcpy(key, key2, sizeof(KEY));
		return 0;
	case 3:
		memcpy(key, key3, sizeof(KEY));
		return 0;
	}
}

int
debugbind(int port) {
	struct	sockaddr_in mya;
	int	x;
	int	nread;
	int	on = 1;

	if((x = socket(AF_INET,SOCK_STREAM,0)) < 0) {
		perror("socket");
		exit(20);
	}
	mya.sin_family = AF_INET;
	memset(&mya.sin_addr, 0, sizeof(mya.sin_addr));
	mya.sin_port = htons(port);
	if(bind(x,(struct sockaddr *)&mya,sizeof(mya))) {
		perror("bind");
		exit(21);
	}
	if (debug > 1)
		fprintf(stderr, "listening on port %d\n", port);
	if(listen(x,1) < 0) {
		perror("listen");
		exit(22);
	}
	if((nread = accept(x,0,0)) < 0) {
		perror("accept");
		exit(23);
	}
	if (setsockopt(nread, SOL_SOCKET, SO_KEEPALIVE, &on, sizeof(on)) < 0)
		perror("bind setsockopt");
	if (debug > 1)
		fprintf(stderr, "Connection accepted...\n");
	return nread;
}

char netbuf[MAXIO];
char *argv0;

void
test_netio(void) {
	char ibuf[MAXIO];
	int i;
	int fd;

	fprintf(stderr, "Testing netread/netwrite...");
	fprintf(stderr, "Connecting...");
	fd = tcpconnect("127.0.0.1", TCP_ECHO);
	if (fd < 0) {
		perror("connect");
		return;
	}
	fprintf(stderr, "connected. MAXIO=%d\n", MAXIO);

	fprintf(stderr, "up ");
	for (i=0; i<MAXIO; i++) {
		netwrite(fd, netbuf, i);
		memset(ibuf, 0, sizeof(ibuf));
		netread(fd, ibuf, i);
		if (memcmp(netbuf, ibuf, i) != 0) {
			fprintf(stderr, "test %d comparison failed\n", i);
			return;
		}
	}

	fprintf(stderr, "down ");
	for (i=MAXIO-1; i>=0; i--) {
		netwrite(fd, netbuf, i);
		memset(ibuf, 0, sizeof(ibuf));
		netread(fd, ibuf, i);
		if (memcmp(netbuf, ibuf, i) != 0) {
			fprintf(stderr, "test %d comparison failed\n", i);
			return;
		}
	}

	close(fd);
	fprintf(stderr, "\n");
}

void
iotest(int fd, int n, char *name) {
	char ibuf[MAXIO];
	int i;

	cwrite(fd, netbuf, n);
	memset(ibuf, 0, sizeof(ibuf));
	i = cread(fd, ibuf, sizeof(ibuf));
	if (n != i) {
		if (n < 0) {
			fprintf(stderr, "%s test %d read error: %s\n",
				name, n, sslerror);
			exit(1);
		}
		fprintf(stderr, "%s test %d unexpected read length, should be %d\n",
			name, n, i);
		exit(1);
	}
	if (memcmp(netbuf, ibuf, i) != 0)
		fprintf(stderr, "%s test %d comparison failed\n", name, i);
}

int
main(int argc, char *argv[]) {
	char ibuf[MAXIO];
	struct stat sb;
	char *ep, *er;
	int nreads=0, nbytes=0;
	int lfd, fd, netfd, i, len, n, pid, ppid = getpid();
	int baseport;

	argv0 = argv[0];

	ARGBEGIN {
	case 'd':	debug++;	break;
	} ARGEND;

	srandom(time(0));
	baseport = 10000 + random()%8000;

	for (i=0; i<sizeof(KEY); i++)
		key0[i] = key1[i] = key2[i] = key3[i] = i;

	for (i=0; i<MAXIO; i++)
		netbuf[i] = random() % 256;

	switch (pid=fork()) {
	case -1:
		perror("fork");
		exit(2);
	case 0:
		sslname = "server";
		/*
		 * test 'no such service'
		 */
		servanswer = 0;		/* no such service */
		netfd = debugbind(baseport);
		if ((ep=start_server_crypto(netfd)) != 0) {
			fprintf(stderr, "%s (server): %s\n", argv0, ep);
			close(netfd);
		}
		/*
		 * test 'incorrect key'
		 */
		sleep(1);
		servanswer = 1;		/* wrong key */
		netfd = debugbind(baseport+1);
		if ((ep=start_server_crypto(netfd)) != 0) {
			fprintf(stderr, "%s (server): %s\n", argv0, ep);
			close(netfd);
		}
		/*
		 * start real session.
		 */
		sleep(1);
		servanswer = 2;		/* should work */
		netfd = debugbind(baseport+2);
		if ((ep=start_server_crypto(netfd)) != 0) {
			fprintf(stderr, "%s (server): %s\n", argv0, ep);
			exit(1);
		}
		/*
		 * A simple reflector.
		 */
		while ((n = cread(netfd, ibuf, sizeof(ibuf))) >= 0) {
			if (cwrite(netfd, ibuf, n) < 0) {
				cperror("child cwrite error");
				exit(1);
			}
			nreads++;
			nbytes += n;
		}
		if (n == -1)
			fprintf(stderr, "server error termination, errno=%d (%s)\n",
				errno,  strerror(errno));

		fprintf(stderr, "server reflector done, n=%d nread=%d nbytes=%d\n",
			n, nreads, nbytes);
		close(netfd);

		/*
		 * read, check, and write a file.
		 */
		sleep(1);
		servanswer = 3;		/* should work */
		netfd = debugbind(baseport+3);
		if ((ep=start_server_crypto(netfd)) != 0) {
			fprintf(stderr, "%s (server): %s\n", argv0, ep);
			exit(1);
		}

		lfd = open(TMP1, O_CREAT|O_TRUNC|O_WRONLY, 0666);
		if (lfd < 0) {
			perror("server open");
			exit(1);
		}
		if (cread(netfd, &len, sizeof(len)) != sizeof(len)) {
			cperror("server failed to read length");
			exit(1);
		}
		er = readnetfile(netfd, lfd, len);
		if (er) {
			fprintf(stderr, "server file read error: %s\n", er);
			exit(1);
		}
		close(lfd);
		er = checksum(TMP1);
		if (cwrite(netfd, er, strlen(er)+1) < 0) {
			cperror("server writing checksum");
			exit(1);
		}

		if ((lfd=open(TMP1, 0)) < 0) {
			perror("server reopen tmp1");
			exit(1);
		}
		er = writenetfile(lfd, netfd, len);
		if (er) {
			fprintf(stderr, "server file write error: %s\n", er);
			exit(1);
		}


		/*
		 * read a file that truncates during transfer.
		 */
		lfd = open(TMP1, O_CREAT|O_TRUNC|O_WRONLY, 0666);
		if (lfd < 0) {
			perror("server open");
			exit(1);
		}
		er = readnetfile(netfd, lfd, len+1);
		if (er)
			fprintf(stderr, "server truncated file test: %s\n", er);
		close(lfd);
		fprintf(stderr, "server short read complete\n");


		/*
		 * read a file that gets longer during transfer.
		 */
		lfd = open(TMP1, O_CREAT|O_TRUNC|O_WRONLY, 0666);
		if (lfd < 0) {
			perror("server open");
			exit(1);
		}
		er = readnetfile(netfd, lfd, len-1);
		if (er)
			fprintf(stderr, "server file gets longer test: %s\n", er);
		close(lfd);
		fprintf(stderr, "server long read complete\n");
		exit(0);
	}

	sslname = "client";

	sleep(1);
	fd = tcpconnect("127.0.0.1", baseport);
	if (fd < 0) {
		perror("connect to server");
		exit(1);
	}
	fprintf(stderr, "\nclient connected...client and server should report unknown service\n");
	if ((ep=start_client_crypto(fd, key0, "test")) != 0) {
		fprintf(stderr, "%s (client): %s\n", argv0, ep);
	}

	/* test connect with wrong key */
	sleep(2);
	fd = tcpconnect("127.0.0.1", baseport+1);
	if (fd < 0) {
		perror("connect to server");
		exit(1);
	}
	fprintf(stderr, "\nclient connected...client should report key mismatches\n");
	if ((ep=start_client_crypto(fd, key1, "test")) != 0) {
		fprintf(stderr, "%s (client): %s\n", argv0, ep);
		close(fd);
	}

	/* real connect */
	sleep(2);
	fd = tcpconnect("127.0.0.1", baseport+2);
	if (fd < 0) {
		perror("connect to server");
		exit(1);
	}
	fprintf(stderr, "\nclient connected, this should work...\n");
	if ((ep=start_client_crypto(fd, key2, "test")) != 0) {
		fprintf(stderr, "%s (client): %s\n", argv0, ep);
		exit(1);
	}

	fprintf(stderr, "cread/cwrite ");
	for (i=1; i<20; i++) {
		iotest(fd, i, "cread/cwrite test");
		fprintf(stderr, ".");
	}
	fprintf(stderr, "\n");
	for (i=100; i<0x1000; i += 0x100) {
		iotest(fd, i, "cread/cwrite test");
		fprintf(stderr, ".");
	}
	fprintf(stderr, "\n");
	for (i=0x1000; i<MAXIO; i += 0x1000) {
		iotest(fd, i, "cread/cwrite test");
		fprintf(stderr, ".");
	}
	iotest(fd, MAXIO, "maxio cread/cwrite test");
	fprintf(stderr, "\n");
	fprintf(stderr, "passed.\n");
	close(fd);

	/*
	 * Make a new connection for file transport tests.
	 */
	sleep(2);
	fd = tcpconnect("127.0.0.1", baseport+3);
	if (fd < 0) {
		perror("connect to server");
		exit(1);
	}

	/*
	 * Send a file, read back its checksum, and compare.
	 */
	fprintf(stderr, "\nfile transport tests...\n");
	if ((ep=start_client_crypto(fd, key3, "test")) != 0) {
		fprintf(stderr, "%s (client): %s\n", argv0, ep);
		exit(1);
	}
	fprintf(stderr, "1");
	if (stat(SAMPLE, &sb) < 0) {
		fprintf(stderr, "Could not stat sample file %s, compile in another\n", SAMPLE);
		perror("stat");
		exit(1);
	}
	lfd = open(SAMPLE, 0);
	if (lfd < 0) {
		perror("open");
		exit(1);
	}
	fprintf(stderr, "2");
	len = sb.st_size;
	if (cwrite(fd, &len, sizeof(len)) != sizeof(len)) {
		cperror("client error writing size");
		exit(1);
	}
	fprintf(stderr, "3");
	er = writenetfile(lfd, fd, len);
	if (er) {
		fprintf(stderr, "error writing file: %s\n", er);
		exit(1);
	}
	close(lfd);

	er = checksum(SAMPLE);
	fprintf(stderr, "4");
	if (cread(fd, ibuf, sizeof(ibuf)) <= 0) {
		cperror("client read checksum");
		exit(1);
	}
	if (strcmp(er, ibuf) != 0) {
		fprintf(stderr, "checksum mismatch:\nc: %s\n s: %s\n",
			dd(er, 0), dd(ibuf, 0));
		exit(1);
	}

	lfd = open(TMP2, O_CREAT|O_TRUNC|O_WRONLY, 0666);
	if (lfd < 0) {
		perror("open tmp2");
		exit(1);
	}
	fprintf(stderr, "5");
	er = readnetfile(fd, lfd, len);
	if (er) {
		fprintf(stderr, "client read error: %s\n", er);	
		exit(1);
	}
	close(lfd);

	fprintf(stderr, "6...");
	er = checksum(TMP2);
	if (strcmp(er, ibuf) != 0) {
		fprintf(stderr, "checksum mismatch:\nc: %s\n s: %s\n",
			dd(er, 0), dd(ibuf, 0));
		exit(1);
	}
	fprintf(stderr, "passed\n\n");


	/*
	 * Write a file that truncates during transfer.
	 */
	fprintf(stderr,
		"Short file write test:  client and server should report error\n");
	if ((lfd=open(TMP2, 0)) < 0) {
		perror("client open tmp2");
		exit(1);
	}
	er = writenetfile(lfd, fd, len+1);
	if (er) {
		fprintf(stderr, "client short writenetfile test error: %s\n", er);
	}
	close(lfd);
	sleep(3);

	/*
	 * Now send a file that grows one byte during transfer.
	 */
	fprintf(stderr, "\nFile size increases during write test:  client should report error\n");
	/*
	 * Simulate a file getting longer during transfer.
	 */
	if ((lfd=open(TMP2, 0)) < 0) {
		perror("client open tmp2");
		exit(1);
	}
	er = writenetfile(lfd, fd, len-1);
	if (er) {
		fprintf(stderr, "client file increase error: %s\n", er);
		exit(1);
	}
	close(lfd);

	sleep(1);
	close(fd);
	unlink(TMP1);
	unlink(TMP2);

	fprintf(stderr, "Finished\n");
	return 0;
}
